/*
 * Copyright 2023 teogor (Teodor Grigor)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dev.teogor.ceres.ui.designsystem

import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.PaddingValues
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.sizeIn
import androidx.compose.material3.ButtonDefaults
import androidx.compose.material3.Icon
import androidx.compose.material3.ProvideTextStyle
import androidx.compose.material3.TextButton
import androidx.compose.runtime.Composable
import androidx.compose.ui.Modifier
import androidx.compose.ui.unit.dp
import dev.teogor.ceres.ui.icons.CeresIcons
import dev.teogor.ceres.ui.theme.MaterialTheme

/**
 * Ceres view toggle button with included trailing icon as well as compact and expanded
 * text label content slots.
 *
 * @param expanded Whether the view toggle is currently in expanded mode or compact mode.
 * @param onExpandedChange Called when the user clicks the button and toggles the mode.
 * @param modifier Modifier to be applied to the button.
 * @param enabled Controls the enabled state of the button. When `false`, this button will not be
 * clickable and will appear disabled to accessibility services.
 * @param compactText The text label content to show in expanded mode.
 * @param expandedText The text label content to show in compact mode.
 */
@Composable
fun CeresViewToggleButton(
  expanded: Boolean,
  onExpandedChange: (Boolean) -> Unit,
  modifier: Modifier = Modifier,
  enabled: Boolean = true,
  compactText: @Composable () -> Unit,
  expandedText: @Composable () -> Unit,
) {
  TextButton(
    onClick = { onExpandedChange(!expanded) },
    modifier = modifier,
    enabled = enabled,
    colors = ButtonDefaults.textButtonColors(
      contentColor = MaterialTheme.colorScheme.onBackground,
    ),
    contentPadding = CeresViewToggleDefaults.ViewToggleButtonContentPadding,
  ) {
    CeresViewToggleButtonContent(
      text = if (expanded) expandedText else compactText,
      trailingIcon = {
        Icon(
          imageVector = if (expanded) CeresIcons.ViewDay else CeresIcons.ShortText,
          contentDescription = null,
        )
      },
    )
  }
}

/**
 * Internal Ceres view toggle button content layout for arranging the text label and
 * trailing icon.
 *
 * @param text The button text label content.
 * @param trailingIcon The button trailing icon content. Default is `null` for no trailing icon.
 */
@Composable
private fun CeresViewToggleButtonContent(
  text: @Composable () -> Unit,
  trailingIcon: @Composable (() -> Unit)? = null,
) {
  Box(
    Modifier
      .padding(
        end = if (trailingIcon != null) {
          ButtonDefaults.IconSpacing
        } else {
          0.dp
        },
      ),
  ) {
    ProvideTextStyle(value = MaterialTheme.typography.labelSmall) {
      text()
    }
  }
  if (trailingIcon != null) {
    Box(Modifier.sizeIn(maxHeight = ButtonDefaults.IconSize)) {
      trailingIcon()
    }
  }
}

/**
 * Ceres view toggle default values.
 */
object CeresViewToggleDefaults {
  // TODO: File bug
  // Various default button padding values aren't exposed via ButtonDefaults
  val ViewToggleButtonContentPadding =
    PaddingValues(
      start = 16.dp,
      top = 8.dp,
      end = 12.dp,
      bottom = 8.dp,
    )
}
